#==================================================================
# printdata.py
# A Python script to write data file for PYOMO
# ---Author---
# W. Bukhsh,
# wbukhsh@gmail.com
# OATS
# Copyright (c) 2015 by W Bukhsh, Glasgow, Scotland
# OATS is distributed under the GNU GENERAL PUBLIC LICENSE v3 (see LICENSE file for details).
#==================================================================
import datetime
import math
import sys
deltaT = 1.0
class printdata(object):
    def __init__(self,datfile,data,model,options):
        self.datfile = datfile
        self.data    = data
        self.model   = model
        self.options = options
    def reducedata(self):
        self.data["demand"]      = self.data["demand"].drop(self.data["demand"][self.data["demand"]['stat'] == 0].index.tolist())
        self.data["branch"]      = self.data["branch"].drop(self.data["branch"][self.data["branch"]['stat'] == 0].index.tolist())
        self.data["shunt"]       = self.data["shunt"].drop(self.data["shunt"][self.data["shunt"]['stat'] == 0].index.tolist())
        self.data["storage"]     = self.data["storage"].drop(self.data["storage"][self.data["storage"]['stat'] == 0].index.tolist())
        self.data["transformer"] = self.data["transformer"].drop(self.data["transformer"][self.data["transformer"]['stat'] == 0].index.tolist())
        self.data["wind"]        = self.data["wind"].drop(self.data["wind"][self.data["wind"]['stat'] == 0].index.tolist())
        self.data["generator"]   = self.data["generator"].drop(self.data["generator"][self.data["generator"]['stat'] == 0].index.tolist())
    def printheader(self):
        f = open(self.datfile, 'w')
        #####PRINT HEADER--START
        f.write('#This is Python generated data file for Pyomo model DCLF.py\n')
        f.write('#_author_:W. Bukhsh\n')
        f.write('#Time stamp: '+ str(datetime.datetime.now())+'\n')
        f.close()
    def printkeysets(self):
        f = open(self.datfile, 'a')
        ##===sets===
        #---set of buses---
        f.write('set B:=\n')
        for i in self.data["bus"].index.tolist():
            f.write(str(self.data["bus"]["name"][i])+"\n")
        f.write(';\n')
        #---set of generators---
        f.write('set G:=\n')
        for i in self.data["generator"].index.tolist():
            f.write(str(self.data["generator"]["name"][i])+"\n")
        f.write(';\n')
        #---set of demands---
        f.write('set D:=\n')
        for i in self.data["demand"]["name"].unique():
            f.write(str(i)+"\n")
        f.write(';\n')
        #---set of wind generators---
        if len(self.data["wind"]["name"])!=0:
            f.write('set WIND:=\n')
            for i in self.data["wind"]["name"].unique():
                f.write(str(i)+"\n")
            f.write(';\n')
        #===parameters===
        #---Real power demand---
        f.write('param PD:=\n')
        for i in self.data["demand"].index.tolist():
            f.write(str(self.data["demand"]["name"][i])+" "+str(float(self.data["demand"]["real"][i])/self.data["baseMVA"]["baseMVA"][0])+"\n")
        f.write(';\n')
        f.write('param VOLL:=\n')
        for i in self.data["demand"].index.tolist():
            f.write(str(self.data["demand"]["name"][i])+" "+str(float(self.data["demand"]["VOLL"][i]))+"\n")
        f.write(';\n')
        f.write('param baseMVA:=\n')
        f.write(str(self.data["baseMVA"]["baseMVA"][0])+"\n")
        f.write(';\n')
        f.close()
    def printnetwork(self):
        f = open(self.datfile, 'a')
        f.write('set LE:=\n 1 \n 2;\n')
        #set of transmission lines
        f.write('set L:=\n')
        for i in self.data["branch"].index.tolist():
            f.write(str(self.data["branch"]["name"][i])+"\n")
        f.write(';\n')
        #set of transformers
        if len(self.data["transformer"]["name"])!=0:
            f.write('set TRANSF:= \n')
            for i in self.data["transformer"].index.tolist():
                f.write(str(self.data["transformer"]["name"][i])+"\n")
            f.write(';\n')
        #---set of generator-bus mapping (gen_bus, gen_ind)---
        f.write('set Gbs:=\n')
        for i in self.data["generator"].index.tolist():
            f.write(str(self.data["generator"]["busname"][i]) + " "+str(self.data["generator"]["name"][i])+"\n")
        f.write(';\n')
        #---set of wind generator-bus mapping (windgen_bus, gen_ind)---
        if len(self.data["wind"]["name"])!=0:
            f.write('set Wbs:=\n')
            for i in self.data["wind"].index.tolist():
                f.write(str(self.data["wind"]["busname"][i]) + " "+str(self.data["wind"]["name"][i])+"\n")
            f.write(';\n')
        #---set of demand-bus mapping (demand_bus, demand_ind)---
        f.write('set Dbs:=\n')
        for i in self.data["demand"].index.tolist():
            f.write(str(self.data["demand"]["busname"][i]) + " "+str(self.data["demand"]["name"][i])+"\n")
        f.write(';\n')
        #---set of reference bus---
        f.write('set b0:=\n')
        slackbus = self.data["generator"]["busname"][self.data["generator"]["type"]==3].tolist()
        for i in slackbus:
            f.write(str(i)+""+"\n")
        f.write(';\n')
        #---param defining system topolgy---
        f.write('param A:=\n')
        for i in self.data["branch"].index.tolist():
            f.write(str(self.data["branch"]["name"][i])+" "+"1"+" "+str(self.data["branch"]["from_busname"][i])+"\n")
        for i in self.data["branch"].index.tolist():
            f.write(str(self.data["branch"]["name"][i])+" "+"2"+" "+str(self.data["branch"]["to_busname"][i])+"\n")
        f.write(';\n')
        #---Transformers---
        if len(self.data["transformer"]["name"])!=0:
            f.write('param AT:= \n')
            for i in self.data["transformer"].index.tolist():
                f.write(str(self.data["transformer"]["name"][i])+" "+"1"+" "+str(self.data["transformer"]["from_busname"][i])+"\n")
            for i in self.data["transformer"].index.tolist():
                f.write(str(self.data["transformer"]["name"][i])+" "+"2"+" "+str(self.data["transformer"]["to_busname"][i])+"\n")
            f.write(';\n')
        f.close()
    def printLF(self):
        f = open(self.datfile, 'a')
        #---Real power generation bounds---
        f.write('param PG:=\n')
        for i in self.data["generator"].index.tolist():
            f.write(str(self.data["generator"]["name"][i])+" "+str(float(self.data["generator"]["PG"][i])/self.data["baseMVA"]["baseMVA"][0])+"\n")
        f.write(';\n')
        #---Real power wind generation bounds---
        if len(self.data["wind"]["name"])!=0:
            f.write('param WG:=\n')
            for i in self.data["wind"].index.tolist():
                f.write(str(self.data["wind"]["name"][i])+" "+str(float(self.data["wind"]["PG"][i])/self.data["baseMVA"]["baseMVA"][0])+"\n")
            f.write(';\n')
        if len(self.data["transformer"]["name"])!=0:
            f.write('param Tap:=\n')
            for i in self.data["transformer"].index.tolist():
                f.write(str(self.data["transformer"]["name"][i])+" "+str(float(self.data["transformer"]["TapRatio"][i]))+"\n")
            f.write(';\n')
        f.close()
    def printDC(self):
        f = open(self.datfile, 'a')
        #---Tranmission line chracteristics for DC load flow---
        f.write('param BL:=\n')
        for i in self.data["branch"].index.tolist():
            f.write(str(self.data["branch"]["name"][i])+" "+str(-1/float(self.data["branch"]["x"][i]))+"\n")
        f.write(';\n')
        #---Transformer chracteristics---
        if len(self.data["transformer"]["name"])!=0:
            f.write('param BLT:=\n')
            for i in self.data["transformer"].index.tolist():
                f.write(str(self.data["transformer"]["name"][i])+" "+str(-float(1/self.data["transformer"]["x"][i]))+"\n")
            f.write(';\n')
        f.close()
    def printOPF(self):
        f = open(self.datfile, 'a')
        #---Real power generation bounds---
        f.write('param PGmin:=\n')
        for i in self.data["generator"].index.tolist():
            f.write(str(self.data["generator"]["name"][i])+" "+str(float(self.data["generator"]["PGLB"][i])/self.data["baseMVA"]["baseMVA"][0])+"\n")
        f.write(';\n')
        f.write('param PGmax:=\n')
        for i in self.data["generator"].index.tolist():
            f.write(str(self.data["generator"]["name"][i])+" "+str(float(self.data["generator"]["PGUB"][i])/self.data["baseMVA"]["baseMVA"][0])+"\n")
        f.write(';\n')
        #---Real power wind generation bounds---
        if len(self.data["wind"]["name"])!=0:
            f.write('param WGmin:=\n')
            for i in self.data["wind"].index.tolist():
                f.write(str(self.data["wind"]["name"][i])+" "+str(float(self.data["wind"]["PGLB"][i])/self.data["baseMVA"]["baseMVA"][0])+"\n")
            f.write(';\n')
            f.write('param WGmax:=\n')
            for i in self.data["wind"].index.tolist():
                f.write(str(self.data["wind"]["name"][i])+" "+str(float(self.data["wind"]["PGUB"][i])/self.data["baseMVA"]["baseMVA"][0])+"\n")
            f.write(';\n')
        #---Tranmission line bounds---
        f.write('param SLmax:=\n')
        for i in self.data["branch"].index.tolist():
            f.write(str(self.data["branch"]["name"][i])+" "+str(float(self.data["branch"]["ContinousRating"][i])/self.data["baseMVA"]["baseMVA"][0])+"\n")
        f.write(';\n')
        #---Transformer chracteristics---
        if len(self.data["transformer"]["name"])!=0:
            f.write('param SLmaxT:=\n')
            for i in self.data["transformer"].index.tolist():
                f.write(str(self.data["transformer"]["name"][i])+" "+str(float(self.data["transformer"]["ContinousRating"][i])/self.data["baseMVA"]["baseMVA"][0])+"\n")
            f.write(';\n')
        #---cost data---
        f.write('param c2:=\n')
        for i in self.data["generator"].index.tolist():
            f.write(str(self.data["generator"]["name"][i])+" "+str(float(self.data["generator"]["costc2"][i]))+"\n")
        f.write(';\n')
        f.write('param c1:=\n')
        for i in self.data["generator"].index.tolist():
            f.write(str(self.data["generator"]["name"][i])+" "+str(float(self.data["generator"]["costc1"][i]))+"\n")
        f.write(';\n')
        f.write('param c0:=\n')
        for i in self.data["generator"].index.tolist():
            f.write(str(self.data["generator"]["name"][i])+" "+str(float(self.data["generator"]["costc0"][i]))+"\n")
        f.write(';\n')
        f.close()
    def printDCOPF(self):
        f = open(self.datfile, 'a')
        #---Tranmission line chracteristics---
        f.write('param BL:=\n')
        for i in self.data["branch"].index.tolist():
            f.write(str(self.data["branch"]["name"][i])+" "+str(-1/float(self.data["branch"]["x"][i]))+"\n")
        f.write(';\n')
        #---Transformer chracteristics---
        if len(self.data["transformer"]["name"])!=0:
            f.write('param BLT:=\n')
            for i in self.data["transformer"].index.tolist():
                f.write(str(self.data["transformer"]["name"][i])+" "+str(-float(1/self.data["transformer"]["x"][i]))+"\n")
            f.write(';\n')
        f.close()
    def printAC(self):
        f = open(self.datfile, 'a')

        #set of shunts
        if len(self.data["shunt"]["name"])!=0:
            f.write('set SHUNT:=\n')
            for i in self.data["shunt"].index.tolist():
                f.write(str(self.data["shunt"]["name"][i])+"\n")
            f.write(';\n')
            f.write('set SHUNTbs:=\n')
            for i in self.data["shunt"].index.tolist():
                f.write(str(self.data["shunt"]["busname"][i])+" "+str(self.data["shunt"]["name"][i])+"\n")
            f.write(';\n')
            f.write('param GB:=\n')
            for i in self.data["shunt"].index.tolist():
                f.write(str(self.data["shunt"]["name"][i])+" "+str(float(self.data["shunt"]["GL"][i])/self.data["baseMVA"]["baseMVA"][0])+"\n")
            f.write(';\n')
            f.write('param BB:=\n')
            for i in self.data["shunt"].index.tolist():
                f.write(str(self.data["shunt"]["name"][i])+" "+str(float(self.data["shunt"]["BL"][i])/self.data["baseMVA"]["baseMVA"][0])+"\n")
            f.write(';\n')
        #---Reactive power demand---
        f.write('param QD:=\n')
        for i in self.data["demand"].index.tolist():
            f.write(str(self.data["demand"]["name"][i])+" "+str(float(self.data["demand"]["reactive"][i])/self.data["baseMVA"]["baseMVA"][0])+"\n")
        f.write(';\n')

        f.write('param G11:=\n')
        for i in self.data["branch"].index.tolist():
            f.write(str(self.data["branch"]["name"][i])+" "+str(self.data["branch"]["r"][i]/(self.data["branch"]["r"][i]**2+self.data["branch"]["x"][i]**2))+"\n")
        f.write(';\n')
        f.write('param G12:=\n')
        for i in self.data["branch"].index.tolist():
            f.write(str(self.data["branch"]["name"][i])+" "+str(-self.data["branch"]["r"][i]/(self.data["branch"]["r"][i]**2+self.data["branch"]["x"][i]**2))+"\n")
        f.write(';\n')
        f.write('param G21:=\n')
        for i in self.data["branch"].index.tolist():
            f.write(str(self.data["branch"]["name"][i])+" "+str(-self.data["branch"]["r"][i]/(self.data["branch"]["r"][i]**2+self.data["branch"]["x"][i]**2))+"\n")
        f.write(';\n')
        f.write('param G22:=\n')
        for i in self.data["branch"].index.tolist():
            f.write(str(self.data["branch"]["name"][i])+" "+str(self.data["branch"]["r"][i]/(self.data["branch"]["r"][i]**2+self.data["branch"]["x"][i]**2))+"\n")
        f.write(';\n')
        f.write('param B11:=\n')
        for i in self.data["branch"].index.tolist():
            f.write(str(self.data["branch"]["name"][i])+" "+str(-self.data["branch"]["x"][i]/(self.data["branch"]["r"][i]**2+self.data["branch"]["x"][i]**2)+0.5*self.data["branch"]["b"][i])+"\n")
        f.write(';\n')
        f.write('param B12:=\n')
        for i in self.data["branch"].index.tolist():
            f.write(str(self.data["branch"]["name"][i])+" "+str(self.data["branch"]["x"][i]/(self.data["branch"]["r"][i]**2+self.data["branch"]["x"][i]**2))+"\n")
        f.write(';\n')
        f.write('param B21:=\n')
        for i in self.data["branch"].index.tolist():
            f.write(str(self.data["branch"]["name"][i])+" "+str(self.data["branch"]["x"][i]/(self.data["branch"]["r"][i]**2+self.data["branch"]["x"][i]**2))+"\n")
        f.write(';\n')
        f.write('param B22:=\n')
        for i in self.data["branch"].index.tolist():
            f.write(str(self.data["branch"]["name"][i])+" "+str(-self.data["branch"]["x"][i]/(self.data["branch"]["r"][i]**2+self.data["branch"]["x"][i]**2)+0.5*self.data["branch"]["b"][i])+"\n")
        f.write(';\n')

        #derived transformer parameters
        if len(self.data["transformer"]["name"])!=0:
            f.write('param G11T:=\n')
            for i in self.data["transformer"].index.tolist():
                temp     = self.data["transformer"]["r"][i]/(self.data["transformer"]["r"][i]**2+self.data["transformer"]["x"][i]**2)
                f.write(str(self.data["transformer"]["name"][i])+" "+str(temp/(self.data["transformer"]["TapRatio"][i]**2))+"\n")
            f.write(';\n')
            f.write('param G12T:=\n')
            for i in self.data["transformer"].index.tolist():
                tempG     = self.data["transformer"]["r"][i]/(self.data["transformer"]["r"][i]**2+self.data["transformer"]["x"][i]**2)
                tempB     = -self.data["transformer"]["x"][i]/(self.data["transformer"]["r"][i]**2+self.data["transformer"]["x"][i]**2)
                f.write(str(self.data["transformer"]["name"][i])+" "+\
                str(-1/(self.data["transformer"]["TapRatio"][i])*(tempG*math.cos(self.data["transformer"]["PhaseShift"][i])-\
                tempB*math.sin(self.data["transformer"]["PhaseShift"][i])))+"\n")
            f.write(';\n')
            f.write('param G21T:=\n')
            for i in self.data["transformer"].index.tolist():
                tempG     = self.data["transformer"]["r"][i]/(self.data["transformer"]["r"][i]**2+self.data["transformer"]["x"][i]**2)
                tempB     = -self.data["transformer"]["x"][i]/(self.data["transformer"]["r"][i]**2+self.data["transformer"]["x"][i]**2)
                f.write(str(self.data["transformer"]["name"][i])+" "+\
                str(-1/(self.data["transformer"]["TapRatio"][i])*(tempG*math.cos(self.data["transformer"]["PhaseShift"][i])+\
                tempB*math.sin(self.data["transformer"]["PhaseShift"][i])))+"\n")
            f.write(';\n')

            f.write('param G22T:=\n')
            for i in self.data["transformer"].index.tolist():
                temp     = self.data["transformer"]["r"][i]/(self.data["transformer"]["r"][i]**2+self.data["transformer"]["x"][i]**2)
                f.write(str(self.data["transformer"]["name"][i])+" "+str(temp)+"\n")
            f.write(';\n')
            f.write('param B11T:=\n')
            for i in self.data["transformer"].index.tolist():
                temp     = -self.data["transformer"]["x"][i]/(self.data["transformer"]["r"][i]**2+self.data["transformer"]["x"][i]**2)
                f.write(str(self.data["transformer"]["name"][i])+" "+str(temp/(self.data["transformer"]["TapRatio"][i]**2))+"\n")
            f.write(';\n')
            f.write('param B12T:=\n')
            for i in self.data["transformer"].index.tolist():
                tempG     = self.data["transformer"]["r"][i]/(self.data["transformer"]["r"][i]**2+self.data["transformer"]["x"][i]**2)
                tempB     = -self.data["transformer"]["x"][i]/(self.data["transformer"]["r"][i]**2+self.data["transformer"]["x"][i]**2)
                f.write(str(self.data["transformer"]["name"][i])+" "+\
                str(-1/(self.data["transformer"]["TapRatio"][i])*(tempB*math.cos(self.data["transformer"]["PhaseShift"][i])+\
                tempG*math.sin(self.data["transformer"]["PhaseShift"][i])))+"\n")
            f.write(';\n')
            f.write('param B21T:=\n')
            for i in self.data["transformer"].index.tolist():
                tempG     = self.data["transformer"]["r"][i]/(self.data["transformer"]["r"][i]**2+self.data["transformer"]["x"][i]**2)
                tempB     = -self.data["transformer"]["x"][i]/(self.data["transformer"]["r"][i]**2+self.data["transformer"]["x"][i]**2)
                f.write(str(self.data["transformer"]["name"][i])+" "+\
                str(-1/(self.data["transformer"]["TapRatio"][i])*(-tempG*math.sin(self.data["transformer"]["PhaseShift"][i])+\
                tempB*math.cos(self.data["transformer"]["PhaseShift"][i])))+"\n")
            f.write(';\n')
            f.write('param B22T:=\n')
            for i in self.data["transformer"].index.tolist():
                temp     = -self.data["transformer"]["x"][i]/(self.data["transformer"]["r"][i]**2+self.data["transformer"]["x"][i]**2)
                f.write(str(self.data["transformer"]["name"][i])+" "+str(temp)+"\n")
            f.write(';\n')
        f.close()
    def printACLF(self):
        f = open(self.datfile, 'a')
        Slack     = self.data["generator"][["name"]][self.data["generator"]["type"]==3]
        f.write('set g0:=\n')
        for i in Slack.index.tolist():
            f.write(str(Slack["name"][i])+"\n")
        f.write(';\n')
        DistSlac  = self.data["generator"][["name"]][self.data["generator"]["type"]==2]
        bustransf = self.data["transformer"][["name","from_busname","to_busname"]][self.data["transformer"]["type"]==2]
        if not DistSlac.empty:
            f.write('set DistSlac:=\n')
            for i in DistSlac.index.tolist():
                f.write(str(DistSlac["name"][i])+"\n")
            f.write(';\n')
        if not bustransf.empty:
            Bvolt   = []
            Transf2 = []
            for i in bustransf.index.tolist():
                Transf2.append(bustransf["name"][i])
                frombus = self.data["bus"]["baseKV"][self.data["bus"]["name"]==bustransf["from_busname"][i]].item()
                tobus   = self.data["bus"]["baseKV"][self.data["bus"]["name"]==bustransf["to_busname"][i]].item()
                if frombus > tobus:
                    Bvolt.append(self.data["bus"]["name"][self.data["bus"]["name"]==bustransf["to_busname"][i]].item())
                else:
                    Bvolt.append(self.data["bus"]["name"][self.data["bus"]["name"]==bustransf["from_busname"][i]].item())
            f.write('set Transf2:=\n')
            for i in Transf2:
                f.write(str(i)+"\n")
            f.write(';\n')
            f.write('set Bvolt:=\n')
            for i in Bvolt:
                f.write(str(i)+"\n")
            f.write(';\n')
            f.write('param VTar:=\n')
            for i in Bvolt:
                f.write(str(i)+' '+str(self.data["bus"]["VM"][self.data["bus"]["name"]==i].item())+"\n")
            f.write(';\n')
        #---Voltage targets---
        f.write('param VS:=\n')
        for i in self.data["generator"].index.tolist():
            f.write(str(self.data["generator"]["name"][i])+" "+str(float(self.data["generator"]["VS"][i]))+"\n")
        f.write(';\n')
        if len(self.data["transformer"]["name"])!=0:
            #---Transformer tap bounds---
            f.write('param TapLB:=\n')
            for i in self.data["transformer"].index.tolist():
                f.write(str(self.data["transformer"]["name"][i])+" "+str(float(self.data["transformer"]["TapLB"][i]))+"\n")
            f.write(';\n')
            f.write('param TapUB:=\n')
            for i in self.data["transformer"].index.tolist():
                f.write(str(self.data["transformer"]["name"][i])+" "+str(float(self.data["transformer"]["TapUB"][i]))+"\n")
            f.write(';\n')
            f.write('param bC:=\n')
            for i in self.data["transformer"].index.tolist():
                f.write(str(self.data["transformer"]["name"][i])+" "+str(self.data["transformer"]["b"][i])+"\n")
            f.write(';\n')
            f.write('param g:=\n')
            for i in self.data["transformer"].index.tolist():
                f.write(str(self.data["transformer"]["name"][i])+" "+str(self.data["transformer"]["r"][i]/(self.data["transformer"]["r"][i]**2+self.data["transformer"]["x"][i]**2))+"\n")
            f.write(';\n')
            f.write('param b:=\n')
            for i in self.data["transformer"].index.tolist():
                f.write(str(self.data["transformer"]["name"][i])+" "+str(-self.data["transformer"]["x"][i]/(self.data["transformer"]["r"][i]**2+self.data["transformer"]["x"][i]**2))+"\n")
            f.write(';\n')
        f.close()
    def printACOPF(self):
        f = open(self.datfile, 'a')
        #---Reactive power generation bounds---
        f.write('param QGmin:=\n')
        for i in self.data["generator"].index.tolist():
            f.write(str(self.data["generator"]["name"][i])+" "+str(float(self.data["generator"]["QGLB"][i])/self.data["baseMVA"]["baseMVA"][0])+"\n")
        f.write(';\n')
        f.write('param QGmax:=\n')
        for i in self.data["generator"].index.tolist():
            f.write(str(self.data["generator"]["name"][i])+" "+str(float(self.data["generator"]["QGUB"][i])/self.data["baseMVA"]["baseMVA"][0])+"\n")
        f.write(';\n')
        #---Voltage bounds---
        f.write('param Vmin:=\n')
        for i in self.data["bus"].index.tolist():
            f.write(str(self.data["bus"]["name"][i])+" "+str(self.data["bus"]["VNLB"][i])+"\n")
        f.write(';\n')
        f.write('param Vmax:=\n')
        for i in self.data["bus"].index.tolist():
            f.write(str(self.data["bus"]["name"][i])+" "+str(self.data["bus"]["VNUB"][i])+"\n")
        f.write(';\n')
        f.close()
    def printBM(self):
        f = open(self.datfile, 'a')
        #---Reactive power generation bounds---
        f.write('param PG:=\n')
        for i in self.data["generator"].index.tolist():
            f.write(str(self.data["generator"]["name"][i])+" "+str(float(self.data["generator"]["PG"][i])/self.data["baseMVA"]["baseMVA"][0])+"\n")
        f.write(';\n')
        f.write('param QG:=\n')
        for i in self.data["generator"].index.tolist():
            f.write(str(self.data["generator"]["name"][i])+" "+str(float(self.data["generator"]["QG"][i])/self.data["baseMVA"]["baseMVA"][0])+"\n")
        f.write(';\n')
        f.write('param bid:=\n')
        for i in self.data["generator"].index.tolist():
            f.write(str(self.data["generator"]["name"][i])+" "+str(float(self.data["generator"]["bid"][i]))+"\n")
        f.write(';\n')
        f.write('param offer:=\n')
        for i in self.data["generator"].index.tolist():
            f.write(str(self.data["generator"]["name"][i])+" "+str(float(self.data["generator"]["offer"][i]))+"\n")
        f.write(';\n')


        bustransf = self.data["transformer"][["name","from_busname","to_busname"]][self.data["transformer"]["type"]==2]
        if not bustransf.empty:
            Bvolt   = []
            Transf2 = []
            for i in bustransf.index.tolist():
                Transf2.append(bustransf["name"][i])
                frombus = self.data["bus"]["baseKV"][self.data["bus"]["name"]==bustransf["from_busname"][i]].item()
                tobus   = self.data["bus"]["baseKV"][self.data["bus"]["name"]==bustransf["to_busname"][i]].item()
                if frombus > tobus:
                    Bvolt.append(self.data["bus"]["name"][self.data["bus"]["name"]==bustransf["to_busname"][i]].item())
                else:
                    Bvolt.append(self.data["bus"]["name"][self.data["bus"]["name"]==bustransf["from_busname"][i]].item())
            f.write('set Transf2:=\n')
            for i in Transf2:
                f.write(str(i)+"\n")
            f.write(';\n')
            f.write('set Bvolt:=\n')
            for i in Bvolt:
                f.write(str(i)+"\n")
            f.write(';\n')
            f.write('param VTar:=\n')
            for i in Bvolt:
                f.write(str(i)+' '+str(self.data["bus"]["VM"][self.data["bus"]["name"]==i].item())+"\n")
            f.write(';\n')
        #---Voltage targets---
        if len(self.data["transformer"]["name"])!=0:
            #---Transformer tap bounds---
            f.write('param Tap:=\n')
            for i in self.data["transformer"].index.tolist():
                f.write(str(self.data["transformer"]["name"][i])+" "+str(float(self.data["transformer"]["TapRatio"][i]))+"\n")
            f.write(';\n')
            f.write('param TapLB:=\n')
            for i in self.data["transformer"].index.tolist():
                f.write(str(self.data["transformer"]["name"][i])+" "+str(float(self.data["transformer"]["TapLB"][i]))+"\n")
            f.write(';\n')
            f.write('param TapUB:=\n')
            for i in self.data["transformer"].index.tolist():
                f.write(str(self.data["transformer"]["name"][i])+" "+str(float(self.data["transformer"]["TapUB"][i]))+"\n")
            f.write(';\n')
            f.write('param bC:=\n')
            for i in self.data["transformer"].index.tolist():
                f.write(str(self.data["transformer"]["name"][i])+" "+str(self.data["transformer"]["b"][i])+"\n")
            f.write(';\n')
            f.write('param g:=\n')
            for i in self.data["transformer"].index.tolist():
                f.write(str(self.data["transformer"]["name"][i])+" "+str(self.data["transformer"]["r"][i]/(self.data["transformer"]["r"][i]**2+self.data["transformer"]["x"][i]**2))+"\n")
            f.write(';\n')
            f.write('param b:=\n')
            for i in self.data["transformer"].index.tolist():
                f.write(str(self.data["transformer"]["name"][i])+" "+str(-self.data["transformer"]["x"][i]/(self.data["transformer"]["r"][i]**2+self.data["transformer"]["x"][i]**2))+"\n")
            f.write(';\n')

        f.close()

    def printDCBM(self):
        f = open(self.datfile, 'a')
        #---Reactive power generation bounds---
        f.write('param PG:=\n')
        for i in self.data["generator"].index.tolist():
            f.write(str(self.data["generator"]["name"][i])+" "+str(float(self.data["generator"]["PG"][i])/self.data["baseMVA"]["baseMVA"][0])+"\n")
        f.write(';\n')
        f.write('param bidG:=\n')
        for i in self.data["generator"].index.tolist():
            f.write(str(self.data["generator"]["name"][i])+" "+str(float(self.data["generator"]["bid"][i]))+"\n")
        f.write(';\n')
        f.write('param offerG:=\n')
        for i in self.data["generator"].index.tolist():
            f.write(str(self.data["generator"]["name"][i])+" "+str(float(self.data["generator"]["offer"][i]))+"\n")
        f.write(';\n')
        if len(self.data["wind"]["name"])!=0:
            f.write('param PW:=\n')
            for i in self.data["wind"].index.tolist():
                f.write(str(self.data["wind"]["name"][i])+" "+str(float(self.data["wind"]["PG"][i])/self.data["baseMVA"]["baseMVA"][0])+"\n")
            f.write(';\n')

            f.write('param bidW:=\n')
            for i in self.data["wind"].index.tolist():
                f.write(str(self.data["wind"]["name"][i])+" "+str(float(self.data["wind"]["offer"][i]))+"\n")
            f.write(';\n')
        f.close()
    def printUCdat(self):
        f = open(self.datfile, 'a')
        ##===sets===
        f.write('set LE:=\n 1 \n 2;\n')
        #---set of zones---
        f.write('set Z:=\n')
        for i in set(self.data["bus"]["zone"]):
            f.write(str(i)+"\n")
        f.write(';\n')
        #---set of generators---
        f.write('set G:=\n')
        for i in self.data["generator"].index.tolist():
            f.write(str(self.data["generator"]["name"][i])+"\n")
        f.write(';\n')
        #---set of demands---
        f.write('set D:=\n')
        for i in self.data["demand"]["name"].unique():
            f.write(str(i)+"\n")
        f.write(';\n')
        #---set of wind generators---
        if len(self.data["wind"]["name"])!=0:
            f.write('set WIND:=\n')
            for i in self.data["wind"]["name"].unique():
                f.write(str(i)+"\n")
            f.write(';\n')
        #---set of interconnections between zones
        if len(set(self.data["bus"]["zone"]))>1:
            f.write('set ICT:=\n')
            for i in self.data["zonalNTC"].index.tolist():
                f.write(str(self.data["zonalNTC"]["interconnection_ID"][i])+"\n")
            f.write(';\n')
         ###---set of storage---
        if len(self.data["storage"]["name"])!=0:
            f.write('set S:=\n')
            for i in self.data["storage"]["name"].unique():
                f.write(str(i)+"\n")
            f.write(';\n')
        #---set of time-periods---
        f.write('set T:= \n')
        for i in self.data["timeseries"]["Demand"].index.tolist():
            f.write(str(i) + "\n")
        f.write(';\n')
        f.write('set TRed:= \n')
        for i in self.data["timeseries"]["Demand"].index.tolist()[1:]:
            f.write(str(i) + "\n")
        f.write(';\n')
        #---storage-bus mapping---
        if len(self.data["storage"]["name"])!=0:
            f.write('set Sbs:=\n')
            for i in self.data["storage"].index.tolist():
                f.write(str(self.data["storage"]["zone"][i]) + " "+str(self.data["storage"]["name"][i])+"\n")
            f.write(';\n')
        ##---set of GZ(generator zone mapping---
        f.write('set GZ:=\n')
        for i in self.data["generator"].index.tolist():
            f.write(str(self.data["generator"]["name"][i])+" "+str(self.data["bus"]["zone"][self.data["bus"]["name"]== self.data["generator"]["busname"][i]].item())+"\n")
        f.write(';\n')
        ###---set of wind zone mapping---
        if len(self.data["wind"]["name"].index.tolist())!=0:
            f.write('set WZ:=\n')
            for i in self.data["wind"]["name"].index.tolist():
                f.write(str(self.data["wind"]["name"][i])+" "+str(self.data["bus"]["zone"][self.data["bus"]["name"]== self.data["wind"]["busname"][i]].item())+"\n")
            f.write(';\n')
        ###---set of DZ(load zone mapping---
        f.write('set DZ:=\n')
        for i in self.data["demand"]["name"].index.tolist():
            f.write(str(self.data["demand"]["name"][i])+" "+str(self.data["bus"]["zone"][self.data["bus"]["name"]==self.data["demand"]["busname"][i]].item())+"\n")
        f.write(';\n')
        #===parameters===
        #---Real power demand---
        f.write('param PD:=\n')
        for i in self.data["timeseries"]["Demand"]:
            for j in self.data["timeseries"]["Demand"].index.tolist():
                f.write(str(i)+" "+str(j)+" "+str(float(self.data["timeseries"]["Demand"][i][j])/self.data["baseMVA"]["baseMVA"][0])+"\n")
        f.write(';\n')
        f.write('param VOLL:=\n')
        for i in self.data["demand"].index.tolist():
            f.write(str(self.data["demand"]["name"][i])+" "+str(float(self.data["demand"]["VOLL"][i]))+"\n")
        f.write(';\n')
        f.write('param baseMVA:=\n')
        f.write(str(self.data["baseMVA"]["baseMVA"][0])+"\n")
        f.write(';\n')

        #---Real power generation bounds---
        f.write('param PGmin:=\n')
        for i in self.data["generator"].index.tolist():
            f.write(str(self.data["generator"]["name"][i])+" "+str(float(self.data["generator"]["PGLB"][i])/self.data["baseMVA"]["baseMVA"][0])+"\n")
        f.write(';\n')
        f.write('param PGmax:=\n')
        for i in self.data["generator"].index.tolist():
            f.write(str(self.data["generator"]["name"][i])+" "+str(float(self.data["generator"]["PGUB"][i])/self.data["baseMVA"]["baseMVA"][0])+"\n")
        f.write(';\n')
        #---Real power wind generation bounds---
        if len(self.data["wind"]["name"])!=0:
            f.write('param WGmin:=\n')
            for i in self.data["timeseries"]["Wind"]:
                for j in self.data["timeseries"]["Wind"].index.tolist():
                    f.write(str(i)+" "+str(j)+" "+str(0)+"\n")
            f.write(';\n')
            f.write('param WGmax:=\n')
            for i in self.data["timeseries"]["Wind"]:
                for j in self.data["timeseries"]["Wind"].index.tolist():
                    f.write(str(i)+" "+str(j)+" "+str(float(self.data["timeseries"]["Wind"][i][j])/self.data["baseMVA"]["baseMVA"][0])+"\n")
            f.write(';\n')

        if len(self.data["storage"]["name"])!=0:
            f.write('param ChargeEff:=\n')
            for i in  self.data["storage"].index.tolist():
                f.write(str(self.data["storage"]["name"][i])+" "+str(float(self.data["storage"]["ChargingEfficieny(%)"][i])/100.0)+"\n")
            f.write(';\n')
            f.write('param DischargeEff:=\n')
            for i in  self.data["storage"].index.tolist():
                f.write(str(self.data["storage"]["name"][i])+" "+str(float(self.data["storage"]["DischargingEfficieny(%)"][i])/100.0)+"\n")
            f.write(';\n')
            f.write('param StoreUB:=\n')
            for i in  self.data["storage"].index.tolist():
                f.write(str(self.data["storage"]["name"][i])+" "+str(float(self.data["storage"]["capacity(MW)"][i])/self.data["baseMVA"]["baseMVA"][0])+"\n")
            f.write(';\n')
            f.write('param StoreLB:=\n')
            for i in  self.data["storage"].index.tolist():
                f.write(str(self.data["storage"]["name"][i])+" "+str(float(self.data["storage"]["Minoperatingcapacity(MW)"][i])/self.data["baseMVA"]["baseMVA"][0])+"\n")
            f.write(';\n')
            f.write('param StoreInitial:=\n')
            for i in  self.data["storage"].index.tolist():
                f.write(str(self.data["storage"]["name"][i])+" "+str(float(self.data["storage"]["InitialStoredPower(MW)"][i])/self.data["baseMVA"]["baseMVA"][0])+"\n")
            f.write(';\n')
            f.write('param StoreFinal:=\n')
            for i in  self.data["storage"].index.tolist():
                f.write(str(self.data["storage"]["name"][i])+" "+str(float(self.data["storage"]["FinalStoredPower(MW)"][i])/self.data["baseMVA"]["baseMVA"][0])+"\n")
            f.write(';\n')
            f.write('param rampCharge:=\n')
            for i in  self.data["storage"].index.tolist():
                f.write(str(self.data["storage"]["name"][i])+" "+str(float(self.data["storage"]["chargingrate(MW/hr)"][i])/self.data["baseMVA"]["baseMVA"][0])+"\n")
            f.write(';\n')
            f.write('param rampDischarge:=\n')
            for i in  self.data["storage"].index.tolist():
                f.write(str(self.data["storage"]["name"][i])+" "+str(float(self.data["storage"]["dischargingrate(MW/hr)"][i])/self.data["baseMVA"]["baseMVA"][0])+"\n")
            f.write(';\n')


        #--zonal data--
        f.write('param ResReq:=\n')
        for i in self.data["zone"].index.tolist():
            f.write(str(self.data["zone"]["zone"][i])+" "+str(float(self.data["zone"]["reserve(MW)"][i])/self.data["baseMVA"]["baseMVA"][0])+"\n")
        f.write(';\n')
        if len(self.data["zonalNTC"].index.tolist())!=0:
            f.write('param A:=\n')
            for i in self.data["zonalNTC"].index.tolist():
                f.write(str(self.data["zonalNTC"]["interconnection_ID"][i])+" "+"1"+" "+str(self.data["zonalNTC"]["from_zone"][i])+"\n")
            for i in self.data["zonalNTC"].index.tolist():
                f.write(str(self.data["zonalNTC"]["interconnection_ID"][i])+" "+"2"+" "+str(self.data["zonalNTC"]["to_zone"][i])+"\n")
            f.write(';\n')

            f.write('param NTCto:=\n')
            for i in self.data["zonalNTC"].index.tolist():
                f.write(str(self.data["zonalNTC"]["interconnection_ID"][i])+" "+str(float(self.data["zonalNTC"]["TransferCapacityTo(MW)"][i])/self.data["baseMVA"]["baseMVA"][0])+"\n")
            f.write(';\n')
            f.write('param NTCfr:=\n')
            for i in self.data["zonalNTC"].index.tolist():
                f.write(str(self.data["zonalNTC"]["interconnection_ID"][i])+" "+str(float(self.data["zonalNTC"]["TransferCapacityFr(MW)"][i])/self.data["baseMVA"]["baseMVA"][0])+"\n")
            f.write(';\n')

        #---ramp rates---
        f.write('param RampUp:=\n')
        for i in self.data["generator"].index.tolist():
            f.write(str(self.data["generator"]["name"][i])+" "+str(float(deltaT*self.data["generator"]["RampUp(MW/hr)"][i])/self.data["baseMVA"]["baseMVA"][0])+"\n")
        f.write(';\n')
        f.write('param RampDown:=\n')
        for i in self.data["generator"].index.tolist():
            f.write(str(self.data["generator"]["name"][i])+" "+str(float(deltaT*self.data["generator"]["RampDown(MW/hr)"][i])/self.data["baseMVA"]["baseMVA"][0])+"\n")
        f.write(';\n')

        # #---minimum up and down times---
        f.write('param MinUpTime:=\n')
        for i in self.data["generator"].index.tolist():
            f.write(str(self.data["generator"]["name"][i])+" "+str(float((1.0/deltaT)*self.data["generator"]["MinUpTime(hr)"][i]))+"\n")
        f.write(';\n')
        f.write('param MinDownTime:=\n')
        for i in self.data["generator"].index.tolist():
            f.write(str(self.data["generator"]["name"][i])+" "+str(float((1.0/deltaT)*self.data["generator"]["MinDownTime(hr)"][i]))+"\n")
        f.write(';\n')

        #---cost data---
        f.write('param c2:=\n')
        for i in self.data["generator"].index.tolist():
            f.write(str(self.data["generator"]["name"][i])+" "+str(self.data["generator"]["costc2"][i])+"\n")
        f.write(';\n')
        f.write('param c1:=\n')
        for i in self.data["generator"].index.tolist():
            f.write(str(self.data["generator"]["name"][i])+" "+str(self.data["generator"]["costc1"][i])+"\n")
        f.write(';\n')
        f.write('param c0:=\n')
        for i in self.data["generator"].index.tolist():
            f.write(str(self.data["generator"]["name"][i])+" "+str(self.data["generator"]["costc0"][i])+"\n")
        f.write(';\n')
        f.write('param SDcosts:=\n')
        for i in self.data["generator"].index.tolist():
            f.write(str(self.data["generator"]["name"][i])+" "+str(self.data["generator"]["shutdown"][i])+"\n")
        f.write(';\n')
        f.write('param SUcosts:=\n')
        for i in self.data["generator"].index.tolist():
            f.write(str(self.data["generator"]["name"][i])+" "+str(self.data["generator"]["startup"][i])+"\n")
        f.write(';\n')

        f.write('set UCMinUpT:=\n')
        for i in self.data["generator"].index.tolist():
            if int(self.data["generator"]["MinUpTime(hr)"][i])>1:
                for j in range(int(self.data["generator"]["MinUpTime(hr)"][i]),int(self.data["timeseries"]["Demand"].index[-1]+1)):
                    f.write(str(self.data["generator"]["name"][i])+" "+str(j)+" "+str(self.data["timeseries"]["Demand"].index[-1])+"\n")
        f.write(';\n')
        f.write('set UCMinDownT:=\n')
        for i in self.data["generator"].index.tolist():
            if int(self.data["generator"]["MinDownTime(hr)"][i])>1:
                for j in range(int(self.data["generator"]["MinDownTime(hr)"][i]),int(self.data["timeseries"]["Demand"].index[-1]+1)):
                    f.write(str(self.data["generator"]["name"][i])+" "+str(j)+" "+str(self.data["timeseries"]["Demand"].index[-1])+"\n")
        f.write(';\n')
        f.close()
    def printSCdat(self):
        flag_C = 0
        contingencies_id = 1
        contingencies_set = []
        ##--Security constrained data--
        f = open(self.datfile, 'a')
        ##--Generator contingencies--
        for i in self.data["generator"].index.tolist():
            if int(self.data["generator"]["contingency"][i])==1:
                if flag_C==0:
                    f.write('set CG:=\n')
                    flag_C=1
                f.write(str(contingencies_id)+" "+str(self.data["generator"]["name"][i])+"\n")
                contingencies_set.append(contingencies_id)
                contingencies_id += 1
        if flag_C==1:
            f.write(';\n')
        ##--Branch contingencies--
        flag_C = 0
        for i in self.data["branch"].index.tolist():
            if int(self.data["branch"]["contingency"][i])==1:
                if flag_C==0:
                    f.write('set CL:=\n')
                    flag_C=1
                f.write(str(contingencies_id)+" "+str(self.data["branch"]["name"][i])+"\n")
                contingencies_set.append([contingencies_id,str(self.data["branch"]["probability"][i])])
                contingencies_id += 1
        if flag_C==1:
            f.write(';\n')
        ##--Transformer contingencies--
        flag_C = 0
        for i in self.data["transformer"].index.tolist():
            if int(self.data["transformer"]["contingency"][i])==1:
                if flag_C==0:
                    f.write('set CT:=\n')
                    flag_C=1
                f.write(str(contingencies_id)+" "+str(self.data["transformer"]["name"][i])+"\n")
                contingencies_set.append([contingencies_id,str(self.data["transformer"]["probability"][i])])
                contingencies_id += 1
        if flag_C==1:
            f.write(';\n')
        ##--Wind contingencies--
        flag_C = 0
        for i in self.data["wind"].index.tolist():
            if int(self.data["wind"]["contingency"][i])==1:
                if flag_C==0:
                    f.write('set CWIND:=\n')
                    flag_C=1
                f.write(str(contingencies_id)+" "+str(self.data["wind"]["name"][i])+"\n")
                contingencies_set.append([contingencies_id,str(self.data["wind"]["probability"][i])])
                contingencies_id += 1
        if flag_C==1:
            f.write(';\n')
        ##--set of contingencies--
        f.write('set C:=\n')
        for i in contingencies_set:
            f.write(str(i[0])+"\n")
        f.write(';\n')
        if contingencies_set:
            f.write('param probC:=\n')
            for i in contingencies_set:
                f.write(str(i[0])+" "+str(i[1])+"\n")
            f.write(';\n')
        #---ramp rates---
        f.write('param RampUp:=\n')
        for i in self.data["generator"].index.tolist():
            f.write(str(self.data["generator"]["name"][i])+" "+str(float(deltaT*self.data["generator"]["RampUp(MW/hr)"][i])/self.data["baseMVA"]["baseMVA"][0])+"\n")
        f.write(';\n')
        f.write('param RampDown:=\n')
        for i in self.data["generator"].index.tolist():
            f.write(str(self.data["generator"]["name"][i])+" "+str(float(deltaT*self.data["generator"]["RampDown(MW/hr)"][i])/self.data["baseMVA"]["baseMVA"][0])+"\n")
        f.write(';\n')

        f.close()



#==================================================================
# SCOPF.mod
# PYOMO model file of "DC" Security constrained-optimal power flow problem (SCOPF)
# This formulation uses the standard "DC" model of AC power flow equations
# ---Author---
# W. Bukhsh,
# wbukhsh@gmail.com
# OATS
# Copyright (c) 2017 by W Bukhsh, Glasgow, Scotland
# OATS is distributed under the GNU GENERAL PUBLIC LICENSE v3. (see LICENSE file for details).
#==================================================================


#==========Import==========
from __future__ import division
from pyomo.environ import *
#==========================

model = AbstractModel()

# --- SETS ---
model.B      = Set()  # set of buses, as a list of positive integers
model.G      = Set()  # set of generators, as a list of positive integers
model.WIND   = Set()  # set of wind generators, as a list of positive integers
model.D      = Set()  # set of demands, as a list of positive integers
model.L      = Set()  # set of lines, as a list of positive integers
model.TRANSF = Set()  # set of transformers, as a list of positive integers
model.SHUNT  = Set()  # set of shunts, as a list of positive integers
model.LE     = Set()  # line-to and from ends set (1,2)
model.b0     = Set(within=model.B)  # set of reference buses
model.C      = Set()  # set of contigencies

# set of contingencies
model.CL     = Set(within=model.C * model.L)# set of line contigencies
model.CG     = Set(within=model.C * model.G)# set of generator contigencies
model.CT     = Set(within=model.C * model.TRANSF)# set of transformer contigencies
model.CWIND  = Set(within=model.C * model.WIND)# set of wind generator contigencies

# generators, buses, loads linked to each bus b
model.Gbs     = Set(within=model.B * model.G)    # set of generator-node mapping
model.Wbs     = Set(within=model.B * model.WIND) # set of wing-node mapping
model.Dbs     = Set(within=model.B * model.D)    # set of demand-node mapping
model.SHUNTbs = Set(within=model.B*model.SHUNT)  # set of shunt-node mapping

# --- parameters ---
# line matrix
model.A     = Param(model.L*model.LE)       # bus-line (node-arc) matrix
model.AT    = Param(model.TRANSF*model.LE)  # bus-transformer (node-arc) matrix

# demands
model.PD      = Param(model.D, within=Reals)  # real power demand at load d, p.u.
model.VOLL    = Param(model.D, within=Reals)
# generators
model.PGmax    = Param(model.G, within=NonNegativeReals)# max real power of generator, p.u.
model.PGmin    = Param(model.G, within=Reals)# min real power of generator, p.u.
model.WGmax    = Param(model.WIND, within=NonNegativeReals)# max real power of wind generator, p.u.
model.WGmin    = Param(model.WIND, within=NonNegativeReals)# min real power of wind generator, p.u.
model.RampUp   = Param(model.G, within=NonNegativeReals) # ramp up of generator g, p.u.
model.RampDown = Param(model.G, within=NonNegativeReals) # ramp down of generator g, p.u.

# lines and transformer chracteristics and ratings
model.SLmax  = Param(model.L, within=NonNegativeReals) # max real power limit on flow in a line, p.u.
model.SLmaxT = Param(model.TRANSF, within=NonNegativeReals) # max real power limit on flow in line l, p.u.
model.BL     = Param(model.L, within=Reals)  # susceptance of a line, p.u.
model.BLT    = Param(model.TRANSF, within=Reals)  # susceptance of line l, p.u.
#emergency ratings
model.SLmax_E  = Param(model.L, within=NonNegativeReals) # max emergency real power limit on flow in a line, p.u.
model.SLmaxT_E = Param(model.TRANSF, within=NonNegativeReals) # max emergency real power limit on flow in a line, p.u.

# shunt
model.GB = Param(model.SHUNT, within=Reals) #  shunt conductance
model.BB = Param(model.SHUNT, within=Reals) #  shunt susceptance

# cost data
model.c2    = Param(model.G, within=NonNegativeReals)# generator cost coefficient c2 (*pG^2)
model.c1    = Param(model.G, within=NonNegativeReals)# generator cost coefficient c1 (*pG)
model.c0    = Param(model.G, within=NonNegativeReals)# generator cost coefficient c0

model.baseMVA = Param(within=NonNegativeReals)# base MVA

#constants
model.eps = Param(within=NonNegativeReals)
model.nT = Param(within=NonNegativeReals) #total time horizon
model.probC = Param(model.C, domain=NonNegativeReals) #total time horizon

# === Pre-contigency variables ===
# --- control variables ---
model.pG    = Var(model.G, domain= Reals)  #real power generation
model.pW    = Var(model.WIND, domain= Reals) #real power generation from wind
model.pD    = Var(model.D, domain= Reals) #real power demand delivered
model.alpha = Var(model.D, domain= NonNegativeReals) #propotion of real power demand delivered
# --- state variables ---
model.deltaL = Var(model.L, domain= Reals) # angle difference across lines
model.deltaLT = Var(model.TRANSF, domain= Reals) # angle difference across transformers
model.delta = Var(model.B, domain= Reals, initialize=0.0) # voltage phase angle at bus b, rad
model.pL = Var(model.L, domain= Reals) # real power injected at b onto line l, p.u.
model.pLT = Var(model.TRANSF, domain= Reals) # real power injected at b onto transformer line l, p.u.

# === Post-contigency variables ===
# --- control variables ---
model.pGC          = Var(model.G,model.C, domain= Reals)  #post-fault generation set point
model.DeltapG      = Var(model.G,model.C, domain= Reals)  #change in pre-fault generation set point following a contingency
model.DeltapGUp    = Var(model.G,model.C, domain= NonNegativeReals)  #positive change in pre-fault generation set point
model.DeltapGDown  = Var(model.G,model.C, domain= NonNegativeReals)  #negative change in pre-fault generation set point

model.pWC         = Var(model.WIND,model.C, domain= Reals) #post-fault wind generation
model.DeltapW     = Var(model.WIND,model.C, domain= Reals) #change in pre-fault generation set point following a contingency
model.DeltapWDown = Var(model.WIND,model.C, domain= NonNegativeReals) #negative change in pre-fault generation set point(wind can only move downwards)

model.pDC    = Var(model.D,model.C, domain= Reals) #real power demand delivered post-contingency
model.alphaC = Var(model.D,model.C, domain= NonNegativeReals) #propotion of real power demand delivered post-contingency

# --- state variables ---
model.pLC      = Var(model.L,model.C, domain= Reals) # real power injected at b onto line l, p.u.
model.pLTC     = Var(model.TRANSF,model.C, domain= Reals) # real power injected at b onto transformer l, p.u.
model.deltaLC  = Var(model.L,model.C, domain= Reals) # angle difference across lines
model.deltaLTC = Var(model.TRANSF,model.C, domain= Reals) # angle difference across transformers
model.deltaC   = Var(model.B,model.C, domain= Reals, initialize=0.0) # voltage phase angle at bus b, rad

# --- pre- and post- contingency costs
model.FPreCont  = Var() # Objective function component for pre-contingency operation
model.FPostCont = Var(model.C) # Objective function component for post-contingency operation
# --- cost function ---
def objective(model):
    obj = sum(model.probC[c]*model.FPostCont[c] for c in model.C) +(1-sum(model.probC[c] for c in model.C))*model.FPreCont
    return obj
model.OBJ = Objective(rule=objective, sense=minimize)

# --- cost components of the objective function ---
def precontingency_cost(model):
    return model.FPreCont == sum(model.c2[g]*(model.baseMVA*model.pG[g])**2+model.c1[g]*model.baseMVA*model.pG[g]+model.c0[g] for g in model.G)\
    +sum(model.baseMVA*model.VOLL[d]*(1-model.alpha[d])*model.PD[d] for d in model.D)
model.precontingency_cost_const = Constraint(rule=precontingency_cost)

def postcontingency_cost(model,c):
    return model.FPostCont[c] == sum(model.baseMVA*(model.c1[g]*model.DeltapGDown[g,c]+model.c1[g]*model.DeltapGUp[g,c]) for g in model.G)\
    +sum(model.baseMVA*(model.c1[w]*model.DeltapWDown[w,c]) for w in model.WIND)\
    +sum(model.baseMVA*(model.VOLL[d]*(1-model.alphaC[d,c])*model.PD[d]) for d in model.D)
model.postcontingency_cost_const = Constraint(model.C, rule=postcontingency_cost)


# --- Kirchoff's current law Definition at each bus b ---
def KCL_def(model, b):
    return sum(model.pG[g] for g in model.G if (b,g) in model.Gbs) +\
    sum(model.pW[w] for w in model.WIND if (b,w) in model.Wbs) == \
    sum(model.pD[d] for d in model.D if (b,d) in model.Dbs)+\
    sum(model.pL[l] for l in model.L if model.A[l,1]==b)- \
    sum(model.pL[l] for l in model.L if model.A[l,2]==b)+\
    sum(model.pLT[l] for l in model.TRANSF if model.AT[l,1]==b)- \
    sum(model.pLT[l] for l in model.TRANSF if model.AT[l,2]==b)+\
    sum(model.GB[s] for s in model.SHUNT if (b,s) in model.SHUNTbs)
# the next line creates one KCL constraint for each bus
model.KCL_const = Constraint(model.B, rule=KCL_def)

# --- Kirchoff's voltage law on each line and transformer---
def KVL_line_def(model,l):
    return model.pL[l] == (-model.BL[l])*model.deltaL[l]
def KVL_trans_def(model,l):
    return model.pLT[l] == (-model.BLT[l])*model.deltaLT[l]
#the next two lines creates KVL constraints for each line and transformer, respectively.
model.KVL_line_const     = Constraint(model.L, rule=KVL_line_def)
model.KVL_trans_const    = Constraint(model.TRANSF, rule=KVL_trans_def)

# --- demand model ---
def demand_model(model,d):
    return model.pD[d] == model.alpha[d]*model.PD[d]
def demand_LS_bound_Max(model,d):
    return model.alpha[d] <= 1
#the next two lines creates constraints for demand model
model.demandmodelC = Constraint(model.D, rule=demand_model)
model.demandalphaC = Constraint(model.D, rule=demand_LS_bound_Max)

# --- generator power limits ---
def Real_Power_Max(model,g):
    return model.pG[g] <= model.PGmax[g]
def Real_Power_Min(model,g):
    return model.pG[g] >= model.PGmin[g]
#the next two lines creates generation bounds for each generator.
model.PGmaxC = Constraint(model.G, rule=Real_Power_Max)
model.PGminC = Constraint(model.G, rule=Real_Power_Min)

# ---wind generator power limits ---
def Wind_Real_Power_Max(model,w):
    return model.pW[w] <= model.WGmax[w]
def Wind_Real_Power_Min(model,w):
    return model.pW[w] >= model.WGmin[w]
#the next two lines creates generation bounds for each generator.
model.WGmaxC = Constraint(model.WIND, rule=Wind_Real_Power_Max)
model.WGminC = Constraint(model.WIND, rule=Wind_Real_Power_Min)

# --- line power limits ---
def line_lim1_def(model,l):
    return model.pL[l] <= model.SLmax[l]
def line_lim2_def(model,l):
    return model.pL[l] >= -model.SLmax[l]
#the next two lines creates line flow constraints for each line.
model.line_lim1 = Constraint(model.L, rule=line_lim1_def)
model.line_lim2 = Constraint(model.L, rule=line_lim2_def)

# --- power flow limits on transformer lines---
def transf_lim1_def(model,l):
    return model.pLT[l] <= model.SLmaxT[l]
def transf_lim2_def(model,l):
    return model.pLT[l] >= -model.SLmaxT[l]
#the next two lines creates line flow constraints for each transformer.
model.transf_lim1 = Constraint(model.TRANSF, rule=transf_lim1_def)
model.transf_lim2 = Constraint(model.TRANSF, rule=transf_lim2_def)

# --- phase angle constraints ---
def phase_angle_diff1(model,l):
    return model.deltaL[l] == model.delta[model.A[l,1]] - \
    model.delta[model.A[l,2]]
model.phase_diff1 = Constraint(model.L, rule=phase_angle_diff1)

# --- phase angle constraints ---
def phase_angle_diff2(model,l):
    return model.deltaLT[l] == model.delta[model.AT[l,1]] - \
    model.delta[model.AT[l,2]]
model.phase_diff2 = Constraint(model.TRANSF, rule=phase_angle_diff2)

# --- reference bus constraint ---
def ref_bus_def(model,b):
    return model.delta[b]==0
model.refbus = Constraint(model.b0, rule=ref_bus_def)

#======Post-contingency constraints======
# --- Kirchoff's current law Definition at each bus b ---
def KCL_def_PostCnt(model, b,c):
    return sum(model.pGC[g,c] for g in model.G if (b,g) in model.Gbs) +\
    sum(model.pWC[w,c] for w in model.WIND if (b,w) in model.Wbs) == \
    sum(model.pDC[d,c] for d in model.D if (b,d) in model.Dbs)+\
    sum(model.pLC[l,c] for l in model.L if model.A[l,1]==b)- \
    sum(model.pLC[l,c] for l in model.L if model.A[l,2]==b)+\
    sum(model.pLTC[l,c] for l in model.TRANSF if model.AT[l,1]==b)- \
    sum(model.pLTC[l,c] for l in model.TRANSF if model.AT[l,2]==b)+\
    sum(model.GB[s] for s in model.SHUNT if (b,s) in model.SHUNTbs)
# the next line creates one KCL constraint for each bus
model.KCL_const_PostCnt = Constraint(model.B,model.C, rule=KCL_def_PostCnt)

# --- Kirchoff's voltage law on each line ---
def KVL_line_def_PostCnt(model,l,c):
    if (c,l) in model.CL:
        return model.pLC[l,c] == 0
    else:
        return model.pLC[l,c] == (-model.BL[l])*model.deltaLC[l,c]
def KVL_trans_def_PostCnt(model,l,c):
    if (c,l) in model.CT:
        return model.pLTC[l,c] == 0
    else:
        return model.pLTC[l,c] == (-model.BLT[l])*model.deltaLTC[l,c]
#the next two lines create KVL constraints for each line and transformer, respectively.
model.KVL_line_const_PostCnt     = Constraint(model.L,model.C, rule=KVL_line_def_PostCnt)
model.KVL_trans_const_PostCnt    = Constraint(model.TRANSF,model.C, rule=KVL_trans_def_PostCnt)

# --- change in generation set point---
def change_generation_postcont(model,g,c):
    return model.pGC[g,c] == model.pG[g]+model.DeltapG[g,c]
def change_generation_direction_postcont(model,g,c):
    return model.DeltapG[g,c] == model.DeltapGUp[g,c]-model.DeltapGDown[g,c]
#the next two lines creates constraints for modelling the change in generation from pre-contingency operating point
model.change_generation_postcontC = Constraint(model.G,model.C, rule=change_generation_postcont)
model.change_generation_direction_postcontC = Constraint(model.G,model.C, rule=change_generation_direction_postcont)

# --- change in wind generation set point---
def change_windgeneration_postcont(model,w,c):
    return model.pWC[w,c] == model.pW[w]+model.DeltapW[w,c]
def change_windgeneration_direction_postcont(model,w,c):
    return model.DeltapW[w,c] == -model.DeltapWDown[w,c]
#the next two lines creates constraints for modelling the change in wind generation from pre-contingency operating point
model.change_windgeneration_postcontC = Constraint(model.WIND,model.C, rule=change_windgeneration_postcont)
model.change_windgeneration_direction_postcontC = Constraint(model.WIND,model.C, rule=change_windgeneration_direction_postcont)

# --- demand model (post-contingency)---
def demand_model_postcont(model,d,c):
    return model.pDC[d,c] == model.alphaC[d,c]*model.PD[d]
def demand_LS_bound_Max_postcont(model,d,c):
    return model.alphaC[d,c] <= 1
#the next two lines creates constraints for demand model
model.demandmodel_postcontC = Constraint(model.D,model.C, rule=demand_model_postcont)
model.demandalpha_postcontC = Constraint(model.D,model.C, rule=demand_LS_bound_Max_postcont)

# --- generator power limits ---
def Real_Power_Max_PostCnt(model,g,c):
    if (c,g) in model.CG:
        return model.pGC[g,c] == 0
    else:
        return model.pGC[g,c] <= model.PGmax[g]
def Real_Power_Min_PostCnt(model,g,c):
    if (c,g) in model.CG:
        return model.pGC[g,c] == 0
    else:
        return model.pGC[g,c] >= model.PGmin[g]
#the next two lines create constraints on real power generation bounds
model.PGmaxC_PostCnt = Constraint(model.G,model.C, rule=Real_Power_Max_PostCnt)
model.PGminC_PostCnt = Constraint(model.G,model.C, rule=Real_Power_Min_PostCnt)

# --- wind generator power limits ---
def Wind_Real_Power_Max_PostCnt(model,w,c):
    if (c,w) in model.CWIND:
        return model.pWC[w,c] == 0
    else:
        return model.pWC[w,c] <= model.WGmax[w]
def Wind_Real_Power_Min_PostCnt(model,w,c):
    if (c,w) in model.CWIND:
        return model.pWC[w,c] == 0
    else:
        return model.pWC[w,c] >= model.WGmin[w]
#the next two lines create constraints on real power generation bounds
model.WPGmaxC_PostCnt = Constraint(model.WIND,model.C, rule=Wind_Real_Power_Max_PostCnt)
model.WPGminC_PostCnt = Constraint(model.WIND,model.C, rule=Wind_Real_Power_Min_PostCnt)

# --- line power limits ---
def line_lim1_def_PostCnt(model,l,c):
    return model.pLC[l,c] <= model.SLmax[l]
def line_lim2_def_PostCnt(model,l,c):
    return model.pLC[l,c] >= -model.SLmax[l]
#the next two lines create line flow constraints for each line.
model.line_lim1_PostCnt = Constraint(model.L,model.C, rule=line_lim1_def_PostCnt)
model.line_lim2_PostCnt = Constraint(model.L,model.C, rule=line_lim2_def_PostCnt)

# --- power flow limits on transformer lines---
def transf_lim1_def_PostCnt(model,l,c):
    return model.pLTC[l,c] <= model.SLmaxT[l]
def transf_lim2_def_PostCnt(model,l,c):
    return model.pLTC[l,c] >= -model.SLmaxT[l]
#the next two lines create line flow constraints for each transformer.
model.transf_lim1_PostCnt = Constraint(model.TRANSF,model.C, rule=transf_lim1_def_PostCnt)
model.transf_lim2_PostCnt = Constraint(model.TRANSF,model.C, rule=transf_lim2_def_PostCnt)

# --- phase angle constraints ---
def phase_angle_diff1_PostCnt(model,l,c):
    return model.deltaLC[l,c] == model.deltaC[model.A[l,1],c] - \
    model.deltaC[model.A[l,2],c]
#the next line creates a constraint to link angle difference across a line.
model.phase_diff1_PostCnt = Constraint(model.L,model.C, rule=phase_angle_diff1_PostCnt)

# --- phase angle constraints ---
def phase_angle_diff2_PostCnt(model,l,c):
    return model.deltaLTC[l,c] == model.deltaC[model.AT[l,1],c] - \
    model.deltaC[model.AT[l,2],c]
#the next line creates a constraint to link angle difference across a transformer.
model.phase_diff2_PostCnt = Constraint(model.TRANSF,model.C, rule=phase_angle_diff2_PostCnt)

# --- reference bus constraint ---
def ref_bus_def_PostCnt(model,b,c):
    return model.deltaC[b,c]==0
model.refbus_PostCnt = Constraint(model.b0,model.C, rule=ref_bus_def_PostCnt)

#======Pre-contingency and Post-contingency linking constraints======
def generation_link_rampUp_def(model,g,c):
    if (c,g) in model.CG:
        return model.pGC[g,c] == 0
    else:
        return model.DeltapG[g,c]<=model.RampUp[g]
def generation_link_rampDown_def(model,g,c):
    if (c,g) in model.CG:
        return model.pGC[g,c] == 0
    else:
        return model.DeltapG[g,c]>=-model.RampDown[g]
model.generation_link_rampUp_const   = Constraint(model.G,model.C, rule=generation_link_rampUp_def)
model.generation_link_rampDown_const = Constraint(model.G,model.C, rule=generation_link_rampDown_def)



from setuptools import setup
with open('README.md') as f:
    readme = f.read()
setup(name='oatspower',
      version='0.1.10',
      description='OATS: Optimisation and Analysis Toolbox for Power Systems',
      long_description=readme,
      long_description_content_type='text/markdown',
      url='https://github.com/bukhsh/oats',
      author='Waqquas Bukhsh',
      author_email='wbukhsh@gmail.com',
      license='GPLv3+',
      packages=['oats','oats.models','oats.testcases'],
      package_data={'oats.testcases': ['*']},
      install_requires =[
        'pyomo',
        'tabulate',
        'PyYAML',
        'ipython[notebook]',
        'openpyxl',
        'pymysql',
        'pypyodbc',
        'pyro4',
        'sympy',
        'xlrd',
        'pandas',
      ],
      zip_safe=False)
